use std::sync::Arc;

use axum::{
    extract::{Json, Path},
    http::StatusCode,
    response::{IntoResponse, Response},
    Extension,
};
use tracing::{debug, warn};

use crate::state::ApiState;
use ubisync_lib::{
    api::element::{
        ElementCreateRequest, ElementCreateResponse, ElementGetResponse, ElementRemoveResponse,
        ElementSetRequest, ElementSetResponse,
    },
    types::{AppId, ElementId},
};


pub(super) async fn get(
    Path(id): Path<ElementId>,
    app: Extension<AppId>,
    s: Extension<Arc<ApiState>>,
) -> Response {
    let element = s.get_element(id, app.0);
    match element {
        Ok(el) => (
            StatusCode::OK,
            Json {
                0: ElementGetResponse { element: el },
            },
        )
            .into_response(),
        Err(e) => {
            warn!("Could not get element: {}", e);
            StatusCode::NOT_FOUND.into_response()
        }
    }
}

pub(super) async fn create(
    app: Extension<AppId>,
    s: Extension<Arc<ApiState>>,
    Json(req): Json<ElementCreateRequest>,
) -> Response {
    let pot_id = match req.pot {
        Some(p) => p,
        None => match s.get_default_pot(app.0) {
            Ok(Some(p)) => p.id,
            Ok(_) => {
                warn!("Pot not found");
                return StatusCode::INTERNAL_SERVER_ERROR.into_response();
            },
            Err(e) => {
                warn!("Element create request did not provide pot id, and no default pot for requesting app was found: {}", e);
                return StatusCode::INTERNAL_SERVER_ERROR.into_response();
            }
        },
    };

    let element_id = s.create_element(req.content, pot_id);
    debug!("{:?}", element_id);
    match element_id {
        Ok(id) => (
            StatusCode::OK,
            Json {
                0: ElementCreateResponse { id },
            },
        )
            .into_response(),
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    }
}

pub(super) async fn set(
    Path(id): Path<ElementId>,
    app: Extension<AppId>,
    s: Extension<Arc<ApiState>>,
    Json(req): Json<ElementSetRequest>,
) -> Response {
    let res = s.write_element_content(id, app.0, req.content);
    match res {
        Ok(_) => (
            StatusCode::OK,
            Json {
                0: ElementSetResponse,
            },
        )
            .into_response(),
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    }
}

pub(super) async fn remove(Path(id): Path<ElementId>, s: Extension<Arc<ApiState>>) -> Response {
    let res = s.remove_element(id);
    match res {
        Ok(_) => (
            StatusCode::OK,
            Json {
                0: ElementRemoveResponse,
            },
        )
            .into_response(),
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    }
}
