use std::sync::Arc;

use anyhow::bail;
use api::{v0::app::App, Api, ApiBuilder};
use comm::CommHandle;
use config::Config;
use i2p::net::I2pSocketAddr;
use node_events::UbisyncNodeEvent;
use state::{ApiState, CommState, State};

use ubisync_lib::{
    peer::Peer,
    types::{AppId, PeerId, PotId},
};

pub mod api;
pub mod comm;
pub mod config;
pub mod node_events;
pub mod state;

pub struct Ubisync {
    comm_handle: Arc<CommHandle>,
    state_handle: Arc<State>,
    api: Arc<Api>,
}

impl Ubisync {
    pub async fn new(config: &Config) -> anyhow::Result<Self> {
        let state = State::new(config.database_location.clone()).await?;
        let comm_handle = Arc::new(CommHandle::new(CommState::new(state.clone()), config)?);
        state.set_comm_handle(comm_handle.clone());

        let api = Arc::new(
            ApiBuilder::from(config.api_config.clone())
                .build(ApiState::new(state.clone(), &config.jwt_secret))
                .await,
        );

        comm_handle.run().await;
        Ok(Ubisync {
            comm_handle: comm_handle,
            state_handle: state,
            api: api,
        })
    }

    pub fn set_node_event_callback<CallbackFunction>(
        &self,
        cb: CallbackFunction,
        node: Arc<Ubisync>,
    ) where
        CallbackFunction: Fn(UbisyncNodeEvent, Arc<Ubisync>) + Send + Sync + 'static,
    {
        self.state_handle.set_node_event_callback(cb, node);
    }

    pub fn api(&self) -> Arc<Api> {
        self.api.clone()
    }

    pub fn add_peer(&self, p: impl TryInto<Peer, Error = anyhow::Error>) -> anyhow::Result<()> {
        match p.try_into() {
            Ok(peer) => self.state_handle.set_peer(peer),
            Err(e) => bail!(e),
        }
    }

    pub fn add_peer_from_id(&self, id: PeerId) -> anyhow::Result<()> {
        // TODO: resolve peer's name before setting
        self.state_handle.set_peer(Peer::new(id, None))
    }

    pub fn get_apps(&self) -> Vec<App> {
        self.state_handle.get_apps().unwrap_or(vec![])
    }

    pub fn get_peers(&self) -> Vec<Peer> {
        self.state_handle.get_peers().unwrap_or(vec![])
    }

    pub fn get_destination(&self) -> anyhow::Result<I2pSocketAddr> {
        self.comm_handle.i2p_address()
    }

    pub fn add_pot_member(&self, pot: PotId, app: AppId) -> anyhow::Result<()> {
        self.state_handle.add_pot_member(pot, app)
    }
}
