use std::{sync::Arc, time::Duration};

use anyhow::Error;
use jsonwebtoken::{DecodingKey, EncodingKey, Validation};
use tracing::debug;
use ubisync_lib::{
    api::events::AppEvent,
    messages::MessageContent,
    types::{AppId, Element, ElementContent, ElementId, Pot, PotId},
};

use crate::api::v0::app::App;

use super::{database::StateDB, State};

pub struct ApiState {
    state: Arc<State>,
    jwt_encoding_key: EncodingKey,
    jwt_decoding_key: DecodingKey,
    jwt_validation: Validation,
}

impl ApiState {
    pub fn new(state: Arc<State>, jwt_secret: &str) -> Self {
        let mut validation = Validation::default();
        validation.set_required_spec_claims(&vec!["sub"]);
        let api_state = ApiState {
            state: state,
            jwt_encoding_key: EncodingKey::from_secret(jwt_secret.as_bytes()),
            jwt_decoding_key: DecodingKey::from_secret(jwt_secret.as_bytes()),
            jwt_validation: validation,
        };

        api_state
    }

    pub fn add_app(
        &self,
        name: String,
        description: String,
        app_type: String,
    ) -> anyhow::Result<AppId> {
        let id = AppId::new();
        self.db().add_app(id.clone(), name, description, app_type)?;
        debug!("Successfully added app");

        Ok(id)
    }

    pub fn app_exists(&self, id: AppId) -> anyhow::Result<bool> {
        Ok(self.db().get_app(id)?.is_some())
    }

    pub fn get_app(&self, id: AppId) -> anyhow::Result<App> {
        self.db()
            .get_app(id)
            .map(|app_opt| app_opt.ok_or(Error::msg("Failed to find app")))?
    }

    pub fn create_element(&self, content: ElementContent, pot: PotId) -> anyhow::Result<ElementId> {
        let id = ElementId::new();
        self.db()
            .add_element(id.clone(), content.clone(), None, false, pot.clone())?;
        debug!("Added element {{{}}}", id.to_string());

        self.state.send_to_peers(MessageContent::CreateElement {
            id: id.clone(),
            content: content,
            pot: pot,
        });

        Ok(id)
    }

    pub fn write_element_content(
        &self,
        id: ElementId,
        app: AppId,
        content: ElementContent,
    ) -> anyhow::Result<()> {
        if self.db().app_has_access(app, id.clone())? {
            self.db().set_element_content(id.clone(), content)?;
            self.db().set_element_local_changes(id.clone(), true)?;
            debug!("Wrote element content {{{}}}", id.to_string());

            Ok(())
        } else {
            Err(Error::msg(
                "Element does not exist or app does not have access to it",
            ))
        }
    }

    pub fn remove_element(&self, id: ElementId) -> anyhow::Result<()> {
        self.db()
            .remove_element(id.clone())
            .inspect(|_| debug!("Removed element {{{}}}", id.to_string()))
    }

    pub fn set_app_default_pot(&self, app_id: AppId, pot_id: PotId) -> anyhow::Result<()> {
        self.db().set_default_pot(app_id, pot_id)
    }

    pub fn get_default_pot(&self, app_id: AppId) -> anyhow::Result<Option<Pot>> {
        self.db().get_default_pot(app_id)
    }

    pub fn create_pot(&self, app_id: AppId, app_type: String) -> anyhow::Result<PotId> {
        let pot_id = PotId::new();
        self.db().add_pot(pot_id.clone(), app_type.clone())?;
        self.db().add_pot_membership(pot_id.clone(), app_id)?;

        self.state.send_to_peers(MessageContent::AddPot {
            id: pot_id.clone(),
            app_type: app_type,
        });
        Ok(pot_id)
    }

    pub fn get_element(&self, id: ElementId, app: AppId) -> anyhow::Result<Element> {
        if self.db().app_has_access(app, id.clone())? {
            self.db()
                .get_element(id)
                .ok_or(Error::msg("Could not get element"))
        } else {
            Err(Error::msg(
                "Element does not exist or app does not have access to it",
            ))
        }
    }

    pub fn jwt_encoding_key(&self) -> &EncodingKey {
        &self.jwt_encoding_key
    }

    pub fn jwt_decoding_key(&self) -> &DecodingKey {
        &self.jwt_decoding_key
    }

    pub fn jwt_validation(&self) -> &Validation {
        &self.jwt_validation
    }

    pub async fn events(
        &self,
        app: AppId,
        timeout: u16,
        accumulate: u16,
    ) -> anyhow::Result<Vec<AppEvent>> {
        match self.state.get_event_receiver(&app) {
            Ok(rx_mutex) => {
                if let Ok(receiver) = rx_mutex.lock() {
                    let mut events = vec![];
                    loop {
                        //TODO: Use recv_deadline once it's stable
                        match receiver.recv_timeout(Duration::from_secs(timeout.into())) {
                            Ok(ev) => {
                                debug!("ApiState received event {:?}", &ev);
                                events.push(ev);
                                if events.len() >= accumulate.into() {
                                    return Ok(events);
                                }
                            }
                            Err(_) => return Ok(events),
                        }
                    }
                } else {
                    Err(Error::msg("Failed to lock on receiver mutex"))
                }
            }
            Err(e) => Err(e),
        }
    }

    fn db(&self) -> &StateDB {
        &self.state.db
    }
}

#[cfg(test)]
mod tests {
    use tracing::Level;
    use ubisync_lib::types::ElementContent;

    use crate::state::State;

    use super::ApiState;

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_create() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let state = ApiState::new(
            State::new(None).await.unwrap(),
            "abcdabcdabcdabcdabcdabcdabcdabcd",
        );
        let app_id = state
            .add_app(
                "appname".to_string(),
                "appdesc".to_string(),
                "apptype".to_string(),
            )
            .unwrap();
        let pot_id = state
            .create_pot(app_id.clone(), "apptype".to_string())
            .unwrap();
        let id = state
            .create_element(
                ElementContent::Text("Test-text".to_string()),
                pot_id.clone(),
            )
            .unwrap();
        let el = state.get_element(id, app_id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text".to_string()),
            el.content().to_owned()
        )
    }

    #[tokio::test]
    #[serial_test::serial]
    async fn test_element_write() {
        tracing_subscriber::fmt()
            .pretty()
            .with_max_level(Level::DEBUG)
            .init();

        let state = ApiState::new(
            State::new(None).await.unwrap(),
            "abcdabcdabcdabcdabcdabcdabcdabcd",
        );
        let app_id = state
            .add_app(
                "appname".to_string(),
                "appdesc".to_string(),
                "apptype".to_string(),
            )
            .unwrap();
        let pot_id = state
            .create_pot(app_id.clone(), "apptype".to_string())
            .unwrap();
        let id = state
            .create_element(
                ElementContent::Text("Test-text".to_string()),
                pot_id.clone(),
            )
            .unwrap();
        state
            .write_element_content(
                id.clone(),
                app_id.clone(),
                ElementContent::Text("Test-text 2".to_string()),
            )
            .unwrap();
        let el = state.get_element(id, app_id).unwrap();
        assert_eq!(
            ElementContent::Text("Test-text 2".to_string()),
            el.content().to_owned()
        )
    }
}
