use std::{sync::Arc, time::Duration};

use tracing::{debug, warn, Level};
use ubisync::{config::Config, node_events::UbisyncNodeEvent, Ubisync};
use ubisync_lib::{
    api::{
        element::{ElementCreateRequest, ElementGetRequest},
        events::PollEventsRequest,
    },
    types::{Element, ElementContent},
};
use ubisync_sdk::UbisyncClient;

#[tokio::test(flavor = "multi_thread")]
async fn two_nodes_element_creation() {
    tracing_subscriber::fmt()
        .pretty()
        .with_max_level(Level::DEBUG)
        .init();

    // Two nodes need to bind to different ports
    let mut c2 = Config::default();
    c2.api_config.port = Some(9982);
    let ubi1 = Ubisync::new(&Config::default()).await.unwrap();
    let ubi2 = Arc::new(Ubisync::new(&c2).await.unwrap());
    ubi1.add_peer_from_id(ubi2.get_destination().unwrap().into())
        .unwrap();

    let api_client1 =
        UbisyncClient::init("localhost", 9981, None, "App", "Long desc", "test-app-type")
            .await
            .unwrap()
            .create_default_pot()
            .await
            .unwrap();
    let api_client2 =
        UbisyncClient::init("localhost", 9982, None, "App", "Long desc", "test-app-type")
            .await
            .unwrap();

    let app_id2 = api_client2.app_id();
    ubi2.set_node_event_callback(
        move |ev, node| {
            if let UbisyncNodeEvent::NewPot { id, app_type } = ev {
                debug!("callback called");
                node.add_pot_member(id, app_id2.clone()).unwrap();
            }
        },
        ubi2.clone(),
    );

    tokio::time::sleep(Duration::from_millis(5000)).await;

    let test_element_content = ElementContent::Text("Text".to_string());
    let create_resp = api_client1
        .send(
            ElementCreateRequest {
                content: test_element_content.clone(),
                pot: None,
            },
            (),
        )
        .await
        .unwrap();
    let id = create_resp.id;

    tokio::time::sleep(Duration::from_millis(1000)).await;

    let mut get_resp = api_client2.send(ElementGetRequest {}, id.clone()).await;
    while let Err(_) = get_resp {
        warn!("Sleeping for another second, element has not arrived yet");
        tokio::time::sleep(Duration::from_millis(1000)).await;
        get_resp = api_client2.send(ElementGetRequest {}, id.clone()).await;
    }

    let received_element: Element = get_resp.unwrap().element;
    debug!("Other node received this element: {:?}", received_element);

    assert_eq!(&test_element_content, received_element.content());
    std::process::exit(0);
}

#[tokio::test(flavor = "multi_thread")]
async fn two_nodes_api_event() {
    tracing_subscriber::fmt()
        .pretty()
        .with_max_level(Level::DEBUG)
        .init();

    // Two nodes need to bind to different ports
    let mut c2 = Config::default();
    c2.api_config.port = Some(9982);
    let ubi1 = Arc::new(Ubisync::new(&Config::default()).await.unwrap());
    let ubi2 = Ubisync::new(&c2).await.unwrap();
    ubi2.add_peer_from_id(ubi1.get_destination().unwrap().into())
        .unwrap();

    let api_client1 =
        UbisyncClient::init("localhost", 9981, None, "App", "Long desc", "test-app-type")
            .await
            .unwrap();

    let app_id1 = api_client1.app_id();
    ubi1.set_node_event_callback(
        move |ev, node| {
            debug!("callback called");
            if let UbisyncNodeEvent::NewPot { id, app_type } = ev {
                node.add_pot_member(id, app_id1.clone()).unwrap();
            }
        },
        ubi1.clone(),
    );

    tokio::spawn(async move {
        loop {
            if let Ok(events) = api_client1
                .send(
                    PollEventsRequest {
                        timeout: 60,
                        accumulate: 1,
                    },
                    (),
                )
                .await
            {
                if events.events.len() > 0 {
                    debug!(
                        "Received event from local ubisync node: {:?}",
                        events.events.get(0).unwrap()
                    );
                    std::process::exit(0);
                }

                debug!("Still waiting for an event from the local ubisync node");
                tokio::time::sleep(Duration::from_millis(1000)).await;
            }
        }
    });

    let _api_client2 =
        UbisyncClient::init("localhost", 9982, None, "App", "Long desc", "test-app-type")
            .await
            .unwrap()
            .create_default_pot()
            .await
            .unwrap();

    tokio::time::sleep(Duration::from_millis(10000)).await;
}
